package de.ozgcloud.apilib.common.command.grpc;

import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextAttachingInterceptor;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandId;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.OzgCloudCommandStatus;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.command.GrpcGetCommandRequest;
import lombok.RequiredArgsConstructor;
import net.devh.boot.grpc.client.inject.GrpcClient;

@RequiredArgsConstructor
public class GrpcOzgCloudCommandService implements OzgCloudCommandService {

	private static final int WAIT_TIME_MS = 500;
	public static final int DEFAULT_COMMAND_REQUEST_THRESHOLD_MILLIS = 10000;

	@GrpcClient("ozgcloud-command-manager")
	private final CommandServiceBlockingStub commandServiceStub;

	private final CommandMapper mapper;

	private final OzgCloudCallContextProvider contextProvider;

	private final int commandRequestThresholdMillis;

	@Override
	public OzgCloudCommand createAndWaitUntilDone(OzgCloudCommand commandToCreate) {
		return waitUntilDone(create(commandToCreate));
	}

	@Override
	public OzgCloudCommand create(OzgCloudCommand commandToCreate) {
		var created = getCommandServiceStub().createCommand(mapper.buildCreateCommandRequest(commandToCreate));
		return mapper.fromGrpc(created.getCommand());
	}

	OzgCloudCommand waitUntilDone(OzgCloudCommand commandToWaitFor) {
		var command = commandToWaitFor;
		var timeout = System.currentTimeMillis() + commandRequestThresholdMillis;
		while (!command.getStatus().isFinalState() && System.currentTimeMillis() < timeout) {
			synchronized (this) {
				try {
					wait(WAIT_TIME_MS);
					command = reloadCommand(command.getId());
				} catch (InterruptedException e) {
					Thread.currentThread().interrupt();
				}
			}
		}
		verifyCommand(command);
		return command;
	}

	OzgCloudCommand reloadCommand(OzgCloudCommandId commandId) {
		GrpcGetCommandRequest request = GrpcGetCommandRequest.newBuilder().setId(commandId.toString()).build();
		return mapper.fromGrpc(getCommandServiceStub().getCommand(request));
	}

	CommandServiceBlockingStub getCommandServiceStub() {
		return commandServiceStub.withInterceptors(new OzgCloudCallContextAttachingInterceptor(contextProvider));
	}

	void verifyCommand(OzgCloudCommand command) {
		if (OzgCloudCommandStatus.ERROR.equals(command.getStatus())) {
			throw new TechnicalException("Command (id=%s) failed: %s".formatted(command.getId(), command.getErrorMessage()));
		}
	}
}
